////////////////////////////////////////////////////////////
// Computation of classes of elliptic elements given
// fundamental domain.
////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////
// ordering function for lists.
////////////////////////////////////////////////////////////
class_order:=function(a,b);
v:=a[1]-b[1];
if v gt 0 then 
 return 1;
elif v lt 0 then
 return -1;
else
 return 0;
end if;
end function;//of class_order;
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// ramification_classes takes a fundamental domain
// bounded by x=u1, x=u2 and a group g and computes
// the classes of elliptic elements by computing
// the cycles on vertices and then
// adding the angles associated with each cycle.
//
// The function returns a list of elements of the form
// [vertex_number,matrix] where vertex_number is the 
// the number of the vertex with the smallest x coord
// in its cycle. In addition to any elliptic
// or parabolic elements, the list also contains any
// cases where matrix = [1,0,0,1] i.e. the cycle
// is an ordinary point.
//
// Note that it is sometimes necessary to add extra vertices
// corresponding the self intersection of a bounding circle.
// These are labelled by 1/2 integers.  For example
// if a side labeled by 2 is self paired, then since
// the two vertices are labelled 1 and 2 we include a
// extra vertex labelled by 3/2 corresponding to an
// elliptic fixed point of order 2 at the highest
// point of the circle.
//
// Note that vertices labelled by 1/2 integers are not
// included in the list of cycles since
// they are always fixed points.
//
// Note: these are not the same as vertices which have
// to be added to get all side pairings, which are 
// labelled by integers.
////////////////////////////////////////////////////////////
ramification_classes:=function(u1,u2,refined_cover,g);

in_group := make_in_group(g);

pairs,pairing_matrices,intersection_matrices,intersection_points:=pairing(u1,u2,refined_cover,g);

n:=Maximum([x : x in [Maximum(y[1],y[2]) : y in pairs]]);
pmap:=[[1,0,0,1]:i in [1..n]];
for i:=1 to #pairs do
 m:=pairing_matrices[i];
 mi:=[m[4],-m[2],-m[3],m[1]];
 pmap[pairs[i][1]]:=m;
 pmap[pairs[i][2]]:=mi;
end for;

r_classes:=[[* Rationals()!0,pairing_matrices[1] *]];

for i:=1 to #pairs do
 if pairs[i][1] eq pairs[i][2] then
  Append(~r_classes,[* pairs[i][1]-(1/2),normalize_matrix(pairing_matrices[i]) *]);
 end if;
 if not(in_group(normalize_matrix(pairing_matrices[i]))) then
  print "error in ramification_classes, matrix corresponding to self intersection is not in group",[* (pairs[i][1]-1)+(1/2),normalize_matrix(pairing_matrices[i]) *];
 end if;
end for;

v:=vorbits(pairs);

for j:=1 to #v do
 r:=0;c1:=[1];q1:=[1];s1:=[0];e1:=[1];
  for i in v[j] do
  if i eq 1 then
   c2,q2,s2,e2:=line_angle(refined_cover[1],u1,1);
  elif i eq n-1 then
   c2,q2,s2,e2:=line_angle(refined_cover[n-2],u2,-1);
  else
   c2,q2,s2,e2:=int_angle(refined_cover[i-1],refined_cover[i]);
  end if;
  c1,q1,s1,e1,r1:=angle_add(c1,q1,s1,e1,c2,q2,s2,e2);
  r:=r+r1;
 end for;
 angle_sum:=arc_angle(c1,q1,s1,e1);
 if angle_sum lt 0 then print "error in ramification_classes, cannot compute arc_angle"; end if;
 if angle_sum ge 2 then print "error in ramification_classes, angle_sum is >= 2"; end if;
 if r eq 0 then 
  new_matrix:=[1,0,0,1];
  for i in v[j] do
   new_matrix:=normalize_matrix(prod(pmap[i+1],new_matrix));
  end for;
  Append(~r_classes,[* Minimum(v[j]),new_matrix *]); 
  if not(test_order_and_angle(new_matrix,angle_sum)) then
   print "error, projective order of matrix and sum of angles not consistent in ramification_classes";
   print "matrix =", new_matrix;
   print "angle_sum =", angle_sum;
   print "winding number = ",r;
  end if;
  if not(in_group(new_matrix)) then
   print "error in ramification_classes, computed matrix is not in group, new_matrix =",new_matrix;
  end if;
  old_point := intersection_points[Minimum(v[j])];
  new_point := gact(new_matrix,old_point);
  if new_point ne old_point then
   print "error in ramification_classes, elliptic or parabolic element does not fix expected point, new_matrix, 
    old_point's number, old_point, new_point =", new_matrix, Minimum(v[j]), old_point, new_point;
  end if;
 end if;
 if (r eq 1 and angle_sum eq 0) then
  new_matrix:=[1,0,0,1];
  for i in v[j] do
   new_matrix:=normalize_matrix(prod(pmap[i+1],new_matrix));
  end for;
  Append(~r_classes,[* Minimum(v[j]),new_matrix *]); 
  if new_matrix ne [1,0,0,1] and new_matrix ne [-1,0,0,-1] then
   print "error, projective order of matrix and sum of angles not consistent in ramification_classes";
   print "matrix =", new_matrix;
   print "angle_sum =", angle_sum;
   print "winding number = ",r;
  end if;
 end if;
 if (r eq 1 and angle_sum ne 0) or (r gt 1)  then print "error in ramification_classes, r, angle_sum = ",r, angle_sum;end if;
end for;

Sort(~r_classes,class_order);

//return [y[2] : y in r_classes];
return  r_classes;

end function;// of ramification classes
////////////////////////////////////////////////////////////
