////////////////////////////////////////////////////////////
// functions for matrix products and normalization and
// finding intersection points of two circles.
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// compute product of two matrices without setting up
// group structure
////////////////////////////////////////////////////////////
prod:=function(m1,m2);

a1:=m1[1];
b1:=m1[2];
c1:=m1[3];
d1:=m1[4];
a2:=m2[1];
b2:=m2[2];
c2:=m2[3];
d2:=m2[4];

return( [a1*a2+b1*c2,a1*b2+b1*d2,c1*a2+d1*c2,c1*b2+d1*d2] );

end function;//of prod
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// put matrix into normal form.
////////////////////////////////////////////////////////////
normalize_matrix:=function(n);

den:=[1];
for i:=1 to 4 do
 if n[i] in Rationals() and not(n[i] in Integers()) then 
  Append(~den,Denominator(n[i])); 
 else
  if not(n[i] in Integers()) then
   print "error in normalize_matrix, matrix must have integer or rational entries";
  end if;
 end if;
end for;

lcm:=LeastCommonMultiple(den);

a:=Integers()!(n[1]*lcm);
b:=Integers()!(n[2]*lcm);
c:=Integers()!(n[3]*lcm);
d:=Integers()!(n[4]*lcm);

gcd:=GreatestCommonDivisor([a,b,c,d]);

a:=Integers()!(a/gcd);
b:=Integers()!(b/gcd);
c:=Integers()!(c/gcd);
d:=Integers()!(d/gcd);

if (a-d) ne 0 then 
 s:=Sign(a-d);
elif a ne 0 then
 s:=Sign(a);
elif c ne 0 then
 s:=Sign(c);
elif b ne 0 then
 s:=Sign(b);
elif d ne 0 then
 s:=Sign(d);
else
 s:=1;
end if;

return([s*a,s*b,s*c,s*d]);

end function;//of normalize_matrix
////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////
// determines if cir1 intersects cir2 and if
// so returns coords of intersection point
// in upper half plane and the product of
// the inversions in cir1 and cir2 (which
// fixes the intersection point). The form
// of cir is [c,d,e] for a circle with
// center -d/c and radius 1/(c*sqrt(e))
////////////////////////////////////////////////////////////
intersect:=function(cir1,cir2);



//initialize variables
c1:=cir1[1];
d1:=cir1[2];
e1:=cir1[3];
c2:=cir2[1];
d2:=cir2[2];
e2:=cir2[3];

//special case: if the two circles are equal
//then by convention the intersection point
//is the highest point and the intersection
//matrix is the identity
if cir1 eq cir2 then
 return true,[-d1/c1,1/(c1*e1),-e1],[1,0,0,1];
end if;

//inversions represented by matrices
//note: there is also a complex conjugation
//to be applied with each matrix, but as
//we are taking product of two inversions
//these cancel.
inv1:=[-d1*c1*e1, 1-d1^2*e1, c1^2*e1, c1*d1*e1];
inv2:=[-d2*c2*e2, 1-d2^2*e2, c2^2*e2, c2*d2*e2];

m:=prod(inv1,inv2);
//print "inv1 =",inv1;
//print "inv2 =",inv2;
//print "m =",m;

a:=Integers()!m[1];
b:=Integers()!m[2];
c:=Integers()!m[3];
d:=Integers()!m[4];

gcd:=GreatestCommonDivisor([a,b,c,d]);

a:=Integers()!(a/gcd);
b:=Integers()!(b/gcd);
c:=Integers()!(c/gcd);
d:=Integers()!(d/gcd);

product_mat:=[a,b,c,d];

if c eq 0 then 
 if d-a ne 0 then
  intersects:=true;
  xi:=b/(d-a);
  yi:=[0,1];
 else
  if b eq 0 then
   print "warning, intersection point is not unique";
   intersects:=true;
   xi:=-d1/c1;
   yi:=[1/(c1*e1),-e1];
  else
   intersects:=false;
  end if;
 end if;
else//c ne 0
 rad:=(d-a)^2+4*b*c;
 if rad gt 0 then
  intersects:=false;
 else
  intersects:=true;
  xi:=(a-d)/(2*c);
  if rad eq 0 then
   yi:=[0,1];
  else
   p,q:=SquareFree(rad);
   yi:=[AbsoluteValue(q/(2*c)),p];
  end if;
 end if;
end if;
 
if intersects then 
 point:=psimp([xi,yi[1],yi[2]]);
 return true,point,product_mat ;
else
 return false, _, _  ;
end if;

end function;//of intersect
////////////////////////////////////////////////////////////

