////////////////////////////////////////////////////////////
// Functions for computing the cycles, vertices 
// and side pairings.
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// compute the cycles of vertices of the fundamental domain
////////////////////////////////////////////////////////////
vorbits:=function(pairs);

if pairs eq [ [ ] ] then return [ [ ] ]; end if;

n:=Maximum([x : x in [Maximum(y[1],y[2]) : y in pairs]]);
map:=[1..n];
for i:=1 to #pairs do
 i1:=pairs[i][1];
 i2:=pairs[i][2];
 if i1 ne 1 then map[i1-1]:=i2; end if;
 if i2 ne 1 then map[i2-1]:=i1; end if;
end for;


orbits:=[ ];
new:=[1];
remaining_vertices:={2..n-1};
prior_orbits := [ ];
while #remaining_vertices ge 1 do
 j:=map[new[#new]];
 if (j in prior_orbits) or (j in new and j ne new[1]) then
  return [ [ ] ];
 end if;
 if j eq new[1] then 
  Append(~orbits,new);
  prior_orbits cat:= new;
  x:=Minimum(remaining_vertices);
  new:=[x];
  Exclude(~remaining_vertices,x);
 else
  Append(~new,j);
  Exclude(~remaining_vertices,j);
 end if;
end while;

Append(~orbits,new);

return orbits;

end function;//of vorbits
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// compute vertices of fundamental domain.
////////////////////////////////////////////////////////////
cover_intersections:=function(u1,u2,refined_cover)
k:=#refined_cover;

intersection_points:=[ ];
intersection_matrices:=[ ];

int_true,p,imat:=vintersect(refined_cover[1],u1);
if not(int_true) then 
 //print "Error in cover_intersections, curves do not intersect at left boundary"; 
 return [ ],[ ];
end if;
Append(~intersection_points,p);
Append(~intersection_matrices,imat);

for i:=1 to k-1 do
 int_true,p,imat:=intersect(refined_cover[i],refined_cover[i+1]);
 if not(int_true) then 
  //print "Error in cover_intersections, curves do not intersect between",i,"and ",i+1;
  return [ ],[ ];
 end if;
 Append(~intersection_points,p);
 Append(~intersection_matrices,imat);
end for;

int_true,p,imat:=vintersect(refined_cover[k],u2);
if not(int_true) then 
 //print "Error in cover_intersections, curves do not intersect at right boundary";
 return [ ],[ ];
end if;
Append(~intersection_points,p);
Append(~intersection_matrices,imat);

return intersection_points,intersection_matrices;

end function;//of cover_intersections
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// correct cover to include vertices with angle pi (i.e.
// not at the intersection of two circles). These are
// needed so that we get a consistent set of side pairings.
// If an extra vertex is required, then the circle it is on
// is repeated in the cover so that intersecting consecutive
// circles gives the set of vertices.
////////////////////////////////////////////////////////////
correct_cover:=function(u1,u2,refined_cover,g);

new_cover:=refined_cover;

working := true;
while working do

 intersection_points,intersection_matrices := cover_intersections(u1,u2,new_cover);
 // if failed to find intersection points then
 // return a failed cover
 if intersection_points eq [ ] then
  return [ [ ] ];
 end if;
 k:=#new_cover;
 paired_edges:={};
 searching := true;

 i:=0;
 while i lt k and searching do
  i+:=1;
  if not( i in paired_edges) then
   p1:=intersection_points[i];
   p2:=intersection_points[i+1];
   cir:=new_cover[i];
   mat:=cir_to_mat(u1,u2,p1,p2,cir,g);
   in_group := make_in_group(g);
   np1:=gact(mat,p1);
   np2:=gact(mat,p2);
   if not(np1 in intersection_points) then
    searching := false;
    on_circles := [ j : j in [1..k] | on(np1,new_cover[j]) ];
    if #on_circles ne 1 then
     // print "Error in correct_cover, np1 on more than one circle, np1, np2 in intersection_points,on_circles,new_cover,intersection_points = ",np1,np2 in intersection_points,on_circles,new_cover,intersection_points;
     return [ [ ] ];
    end if;
    new_circle := new_cover[on_circles[1]];
    // test that new point is highest point on cirle. If not, then fail
    dummy,high_point := intersect(new_circle,new_circle);
    if np1 eq high_point then
     new_new_cover := [new_cover[j] : j in [1..on_circles[1]]];
     new_new_cover cat:= [new_cover[on_circles[1]]];
     // print "added circle ",new_cover[on_circles[1]];
     new_new_cover cat:= [new_cover[j] : j in [on_circles[1]+1..#new_cover]];
     new_cover := new_new_cover;
    else
     return [ [ ] ];
    end if;
   elif not(np2 in intersection_points) then
    searching := false;
    on_circles := [ j : j in [1..k] | on(np2,new_cover[j]) ];
    if #on_circles ne 1 then
     // print "Error in correct_cover, np2 on more than one circle, np2, on_circles = ",np2,on_circles;
     return [ [ ] ];
    end if;
    new_circle := new_cover[on_circles[1]];
    // test that new point is highest point on cirle. If not, then fail
    dummy,high_point := intersect(new_circle,new_circle); 
    if np2 eq high_point then
     new_new_cover := [new_cover[j] : j in [1..on_circles[1]]];
     new_new_cover cat:= [new_cover[on_circles[1]]];
     new_new_cover cat:= [new_cover[j] : j in [on_circles[1]+1..#new_cover]];
     new_cover := new_new_cover;
    else
     return [ [ ] ];
    end if;
   else
    ind1:=Index(intersection_points,np1);
    ind2:=Index(intersection_points,np2);
    if ind2 ne ind1-1 then 
     print "Error in cover_intersections, new points are not adjacent for i =",i;
     return [ [ ] ];
    end if;
    paired_edges:=paired_edges join {i,ind2};
   end if;
  end if;
 end while;
 if searching then working:=false; end if;
end while;

return new_cover;
end function;//of correct_cover
////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////
// compute pairings of sides of fundamental domain.
////////////////////////////////////////////////////////////
pairing:=function(u1,u2,refined_cover,g);

intersection_points,intersection_matrices := cover_intersections(u1,u2,refined_cover);

k:=#refined_cover;
pairings:=[[1,k+2]];
pairing_mats:=[[1,u2-u1,0,1]];
paired_edges:={};

for i:=1 to k do
 if not( i in paired_edges) then
  p1:=intersection_points[i];
  p2:=intersection_points[i+1];
  cir:=refined_cover[i];
  mat:=cir_to_mat(u1,u2,p1,p2,cir,g);
  in_group := make_in_group(g);
  if not(in_group(mat)) then
   print "Error in pairing, matrix in not in the group, i,mat = ",i,mat;
  end if;
  np1:=gact(mat,p1);
  ind1:=Index(intersection_points,np1);
  np2:=gact(mat,p2);
  ind2:=Index(intersection_points,np2);
  if ind2 ne ind1-1 then print "Error in pairing, new points are not adjacent for i,ind1,ind2 =",i,ind1,ind2;return [[ ]],_,_;end if;
  Append(~pairings,[i+1,ind2+1]);
  Append(~pairing_mats,mat);
  paired_edges:=paired_edges join {i,ind2};
 end if;
end for;

return pairings,pairing_mats,intersection_matrices,intersection_points;
end function;//of pairing
////////////////////////////////////////////////////////////



