////////////////////////////////////////////////////////////
// functions for working in Q(sqrt(2),sqrt(3),sqrt(5),...)
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// function for ordering lists
////////////////////////////////////////////////////////////
qord:=function(l,m);

if l[1] gt m[1] then return 1; end if;
if l[1] lt m[1] then return -1; end if;
if l[2] gt m[2] then return 1; end if;
if l[2] lt m[2] then return -1; end if;
return 0;
end function;//of qord;
////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////
// simplify an expression in Q(sqrt(2),sqrt(3),sqrt(5),...)
// the expression is of the form:
// c1*sqrt(e1) + c2*sqrt(e2) + ... + cn*sqrt(en)
// where c = [c1,c2,...,cn]
// and   e = [e1,e2,...,en]
////////////////////////////////////////////////////////////
quadsimp:=function(c,q);

if #c ne #q then print "error, quad lengths not equal in quadsimp";end if;

x:=c;
e:=q;

for i:=1 to #x do
 if not(e[i] in Integers()) then
  n:=Integers()!Numerator(e[i]);
  d:=Integers()!Denominator(e[i]);
  if d lt 0 then
   n:=-n;
   d:=-d;
  end if;
  x[i]:=x[i]/d;
  e[i]:=Integers()!(n*d);
 end if;
  //bug in magma, e in integers, but Squarefree
  //does not see this
 if e[i] in Integers() then ei:=Integers()!e[i]; end if;
 //print "e[i] =, e[i] in Integers()",e[i],e[i] in Integers();
 r,s:=Squarefree(ei);
 x[i]:=x[i]*s;
 e[i]:=r;
end for;


w:=[[e[i],x[i]]:i in [1..#e]];
Sort(~w,qord);
nq:=[w[1][1]];
nc:=[Rationals()!w[1][2]];
j:=1;

for i:=2 to #w do
 if w[i][1] ne w[i-1][1] then 
  if nc[j] ne 0 then
   Append(~nq,w[i][1]);Append(~nc,w[i][2]);j:=j+1;
  else
   nq[j]:=w[i][1];nc[j]:=w[i][2];
  end if;
 else
  nc[j]:=nc[j]+w[i][2];
 end if;
end for;

if j gt 1 and nc[j] eq 0 then Remove(~nc,j);Remove(~nq,j);end if;

if j eq 1 and nc[1] eq 0 then
 return [0],[1];
else
 return nc,nq;
end if;

end function;//of quadsimp
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// add two expressions in Q(sqrt(2),sqrt(3),... )
////////////////////////////////////////////////////////////
quadadd:=function(c1,q1,c2,q2);

if #c1 ne #q1 or #c2 ne #q2 then print "error, quad lengths not equal in quadadd";end if;
c:=c1 cat c2;
q:=q1 cat q2;

c,q:=quadsimp(c,q);

return c,q;

end function;//of quadadd
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// multiply two expressions in Q(sqrt(2),sqrt(3),... )
////////////////////////////////////////////////////////////
quadmult:=function(c1,q1,c2,q2);

if #c1 ne #q1 or #c2 ne #q2 then print "error, quad lengths not equal in quadmult";end if;

c:=[ ];
q:=[ ];

for i:=1 to #c1 do
 for j:=1 to #c2 do
  if q1[i] lt 0 and q2[j] lt 0 then 
   Append(~c,Rationals()!(-c1[i]*c2[j]));
   Append(~q,q1[i]*q2[j]);
  else
   Append(~c,Rationals()!(c1[i]*c2[j]));
   Append(~q,q1[i]*q2[j]);
  end if;
end for;
end for;

//print "c,q = ",c,q;
c,q:=quadsimp(c,q);

return c,q;

end function;//of quadmult
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// compute rational upper and lower bounds for e (integer
// greater than or equal to 1) such that 
// |upperbound-lowerbound| < 10^(-k)
////////////////////////////////////////////////////////////
ubl:=function(e,k);

Q:=Rationals();
ub:=Q!(e);
lb:=Q!(1);
ex:=0;
while Floor(ub*10^k) gt lb*10^k do
 ex+:=1;

 //print "ex =",ex;
 ub:= (ub^2+e)/(2*ub);
 lb:= (2*e*lb)/(e+lb^2);
 // note: this next step rounds the bounds to 10^(-ex). Note
 // absolutely necessary, but if not done then the rational
 // expressions get very large. 
 ub:=-Floor(-ub*10^ex)/10^ex;
 lb:=Floor(lb*10^ex)/10^ex;
end while;

// verify the bounds are correct
if (Rationals()!ub)^2-e lt 0 or (e - (Rationals()!lb)^2) lt 0 then print "error in ubl, u,l not bounds, ub^2,lb^2,e,k",ub^2,lb^2,e,k;end if;

return ub,lb;

end function;//of ubl
////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////
// test to see if an expression in Q(sqrt(2),sqrt(3),....)
// is positive, negative or zero.
////////////////////////////////////////////////////////////
quadptest:=function(c,q);
c1,q1:=quadsimp(c,q);

k:=1;

// make sure only positive square roots.
testq:={x ge 0 : x in q1};
if testq ne {true} then
 print "error in quadptest, only positive quadratics allowed";
 return 0;
end if;

// easy cases first.
// note: we have simplified the expression first, so we can
// only get zero if all coefficients are zero. Not true
// if we don't simplify.
if #c1 eq 1  then
 if c1[1] gt 0 
  then return 1;
 elif c1[1] eq 0 then
  return 0;
 elif c1[1] lt 0 then
  return -1;
 end if;
end if;

// initialize upper and lower lists of rational bounds.
cup:= [ Rationals()!1 : i in [1..#c1]]; clo:=cup;
for i:=1 to #c1 do
 if c1[i] ge 0 then 
  ub,lb := ubl(q1[i],k);
  cup[i]:=ub; 
  clo[i]:=lb; 
 else 
  ub,lb := ubl(q1[i],k);
  cup[i]:=lb; 
  clo[i]:=ub; 
 end if;
end for;

ww:=1;
while true do

 ww:=ww+1;
 if ww mod 1000 eq 0 then print "still working in quadptest, ww=",ww;end if;

 lower_bound:=0;
 upper_bound:=0;

 for i:=1 to #c1 do
  lower_bound:=lower_bound + c1[i]*clo[i];
  upper_bound:=upper_bound + c1[i]*cup[i];
 end for;

 if lower_bound gt 0 then return(1); end if;
 if upper_bound lt 0 then return(-1); end if;

 k:=2*k;
 for i:=1 to #c1 do
  if c1[i] ge 0 then 
   ub,lb := ubl(q1[i],k);
   cup[i]:=ub; 
   clo[i]:=lb; 
  else 
   ub,lb := ubl(q1[i],k);
   cup[i]:=lb; 
   clo[i]:=ub; 
  end if;
 end for;
end while;

print "error, did not find sign in ptest!";
return(0);//should never get here!

end function;//of quadptest
////////////////////////////////////////////////////////////

