////////////////////////////////////////////////////////////
// Given a set of circles, remove those which are
// such that only a finite subset can form part of
// the boundary of a fundamental domain.
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Ordering function for sort.
// Sorts circles by x position
// of centre.
////////////////////////////////////////////////////////////
c_order:=function(cir1,cir2);
c1:=cir1[1];
d1:=cir1[2];
c2:=cir2[1];
d2:=cir2[2];
return (-d1/c1) + (d2/c2);
end function;//of c_order
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// True if circle1 is inside circle 2 i.e. no point of
// circle1 are exterior to circle2.
////////////////////////////////////////////////////////////
inside := function(circle1,circle2);

c1 := circle1[1];
d1 := circle1[2];
e1 := circle1[3];
right_point:= [-d1/c1,1/(c1*e1),e1];
left_point:= [-d1/c1,-1/(c1*e1),e1];

if interior_or_on(right_point,circle2) and interior_or_on(left_point,circle2) then
 return true;
else
 return false;
end if;

end function;//of inside
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Compute the endpoints of the boundary arc of the
// circle cover[index] between x=u1 and x=u2.
////////////////////////////////////////////////////////////
endpoints := function(index,cover,u1,u2);


cir := cover[index];
c1 := cir[1];
d1 := cir[2];
e1 := cir[3];
right_real_intersection:= [-d1/c1,1/(c1*e1),e1];
left_real_intersection:= [-d1/c1,-1/(c1*e1),e1];


// special case, circle is leftmost
if index eq 1 then
 intersects_left_line,intersection_point := vintersect(cir,u1);
 if intersects_left_line then
  left_point := intersection_point;
 else 
  left_point := left_real_intersection;
 end if;
 if #cover eq 1 then
  intersects_right_line,intersection_point := vintersect(cir,u2);
  if intersects_right_line then
   right_point := intersection_point;
  else 
   right_point := right_real_intersection;
  end if;
 else
  intersects_next_circle,intersection_point := intersect(cir,cover[index+1]);
  if intersects_next_circle then
   right_point := intersection_point;
  else
   right_point := right_real_intersection;
  end if;
 end if;
// special case circle is rightmost
// note: #cover is now know to be greater than 1.
elif index eq #cover then
 intersects_previous_circle,intersection_point := intersect(cover[index-1],cir);
 if intersects_previous_circle then
  left_point := intersection_point;
 else
  left_point := left_real_intersection;
 end if;
 intersects_right_line,intersection_point := vintersect(cir,u2);
 if intersects_right_line then
  right_point := intersection_point;
 else 
  right_point := right_real_intersection;
 end if;
// usual case
elif index gt 1 and index lt #cover then
 intersects_previous_circle,intersection_point := intersect(cover[index-1],cir);
 if intersects_previous_circle then
  left_point := intersection_point;
 else
  left_point := left_real_intersection;
 end if;
 intersects_next_circle,intersection_point := intersect(cir,cover[index+1]);
 if intersects_next_circle then
  right_point := intersection_point;
 else
  right_point := right_real_intersection;
 end if;
else
 print "error, index is not in range in endpoints";
end if;

return left_point,right_point;

end function;//of endpoints
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// insert cir into cover at position index
////////////////////////////////////////////////////////////
insert:=procedure(u1,u2,~cover,cir,index);

to_remove := [ ];

for i := 1 to #cover do
 left_point,right_point := endpoints(i,cover,u1,u2);
 if interior_or_on(left_point,cir) and interior_or_on(right_point,cir) then
  Append(~to_remove,i);
 end if;
end for;

if to_remove ne [ ] then
 min_index := Minimum(to_remove);
 max_index := Maximum(to_remove);
 // test consistency
 if to_remove ne [min_index..max_index] then
  print "error in insert, values in to_remove are not consecutive";
 end if;
 cover := [cover[i] : i in [1..min_index-1]] cat [cir] cat [cover[i] : i in [max_index+1..#cover]];
else
 cover := [cover[i] : i in [1..index-1]] cat [cir] cat [cover[i] : i in [index..#cover]];
end if;
 

end procedure;//of insert
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// selection of index i of list such that
// list[j] >= a, for all j >= i
//
// Note: assumes that new_cover has already been
// sorted on the x position of the centres of its
// circles.
////////////////////////////////////////////////////////////
find_index := function(cir,new_cover);

c1 := cir[1];
d1 := cir[2];
centre:= -d1/c1;
answer:=-1;
centres := [-new_cover[i,2]/new_cover[i,1] : i in [1..#new_cover]]; 

fast_ge_select(centre,~centres,~answer);

return answer;

end function;//of find_index
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// refines circle_list so that only the circles which
// could form part of the boundary of a fundamental domain
// are kept.
//
// Note: assumes that cover is already sorted by the x coords
// of the centres of the circles.
////////////////////////////////////////////////////////////
refine_cover:=procedure(u1,u2,~cover,~new_circles);

if #new_circles ne 0 then
 Sort(~new_circles,c_order);
 new_cover := cover;
 if #new_cover eq 0 then new_cover:=[new_circles[1]]; end if;
 //intersections := find_intersections(new_cover);

 for i:=1 to #new_circles do
  cir := new_circles[i];
  possible_new_circle := true;
  // find position of leftmost circle whose
  // centre is to the right or the
  // same as the centre of cir. If no such
  // circle return #new_cover +1;
  index := find_index(cir,new_cover);
  // special case, cir is to the left of all circles in new_cover
  if index eq 1 then
   right_circle := new_cover[index];
   if inside(cir,right_circle) then
    possible_new_circle := false;
   end if;
   if possible_new_circle then insert(u1,u2,~new_cover,cir,index); end if;
  // special case: cir is to the right of all circles in new_cover
  elif index eq #new_cover+1 then
   left_circle := new_cover[index-1];
   if inside(cir,left_circle) then
    possible_new_circle := false;
   end if;
   if possible_new_circle then insert(u1,u2,~new_cover,cir,index); end if;
  // normal case
  else
   left_circle := new_cover[index-1];
   right_circle := new_cover[index];
   if inside(cir,left_circle) or inside(cir,right_circle) then
    possible_new_circle := false;
    else
    left_intersects_right,intersection_point := intersect(left_circle,right_circle);
    if left_intersects_right and not(interior(intersection_point,cir)) then possible_new_circle := false; end if;
   end if;
   if possible_new_circle then insert(u1,u2,~new_cover,cir,index); end if;
  end if; // index loop
 end for;// new_circles loop

cover := new_cover;
end if;

end procedure;//of refine_cover
////////////////////////////////////////////////////////////




